# -*- coding: utf-8 -*-
#  Copyright (C) 2007, 2008 Stewart Adam
#  This file is part of fwbackups.

#  fwbackups is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.

#  fwbackups is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.

#  You should have received a copy of the GNU General Public License
#  along with fwbackups; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""
The part of fwbackup that manipulates the actual files.

Remember to modify fwbackups-runonce.py when adding a backend.
"""
import logging
import os
import os.path
import paramiko
import socket
import tarfile
import time
import traceback
#--
import fwbackups
from fwbackups.i18n import _
from fwbackups.const import *
from fwbackups import config
from fwbackups import shutil_modded
from fwbackups import fwlogger
from fwbackups import remote

class EngineError(Exception):
  """ An Engine error """
  def __init__(self, value):
    self.value = value
  def __str__(self):
    return repr(self.value)

class PermissionError(EngineError):
  """ File permission error"""
  def __init__(self, value):
    self.value = value
  def __str__(self):
    return repr(self.value)

class Common:
  """ Common defs for all operations """
  def __init__(self, logger=None, printToo=True):
    # Initialize a logger if none passed
    if not logger:
      self.logger = fwlogger.getLogger()
    else:
      self.logger = logger
    # Totals: Nothing so far
    self.current = 0
    self.total = 0
    # no pids to kill
    self.pids = [0]
    # Execute commands with this environment.
    self.environment = {'PATH':os.getenv('PATH')}
    self.toCancel = False

  def cancelOperation(self):
    """ Stops the operation dead in its tracks """
    for process in self.pids:
      if process == 0:
        continue
      try:
        self.logger.logmsg('DEBUG', _('Killing process with ID %s') % process)
        fwbackups.kill(process, 9)
      except Exception, error:
        self.logger.logmsg('WARNING', _('Could not kill process %(a)s:%(b)s' % (process, error)))
        continue
    self.toCancel = True

  def ifCancel(self):
    """ Check if we should cancel """
    if self.toCancel:
      self.logger.logmsg('INFO', _('The operation has been cancelled.'))
      raise SystemExit

  def getPIDs(self):
    """ Return a list of PIDs """
    return self.pids

  def getProgress(self):
    """
    Get the progress of the operation
    Returns: [current#, total#]
    """
    return [self.current, self.total]

  def onError(self, func, path, exc_info):
    """
    Handles errors when performing file operations.
    func: function called
    path: The path that was being worked on
    exc_info: Good question! I dunno.
    """
    self.logger.logmsg('WARNING', _('Error handling file `%s\'') % path)
    self.logger.logmsg('DEBUG', _('Output:\n%s') % ('func:%s\nexc_info:%s' % (func, exc_info)))
    pass

  def prepDestDir(self, folder):
    """
    Prepare a destination folder
    `folder': The folder to prepare
    """
    if os.path.exists(folder):
      if not os.path.isdir(folder): # what if the user gave a non-folder?
        self.logger.logmsg('WARNING', _('Destination `%s\' exists but is not a folder! Removing.') % folder)
        shutil_modded.rmtree(path=folder, onerror=self.onError)
    elif not (self.engine == 'rsync' and self.incremental == True): # try to create the folder
      try:
        os.mkdir(folder, 0755)
        self.logger.logmsg('DEBUG', _('Created destination folder `%s\'') % folder)
      except:
        error = _('The destination folder `%s\' could not be created.') % folder
        self.logger.logmsg('ERROR', error)
        return False
    # finally
    if not fwbackups.CheckPerms(folder): # make sure it's writable
      error = _('You do not have read and write permissions on the destination `%s\'' % folder)
      self.logger.logmsg('ERROR', error)
      return False
    return True

  def removeIfExists(self, treefile):
    """
    Removes a folder tree or file if it exists already
    `treefile': The tree or folder to remove if existing
    """
    if not (self.engine == 'rsync' and self.incremental == True):
      if os.path.exists(treefile):
        self.logger.logmsg('WARNING', _('`%s\' exists and will be overwritten.') % treefile)
        shutil_modded.rmtree(path=ConvertPath(treefile), onerror=self.onError)
    return True

class Restore(Common):
  """ The one-time backup operation """
  def __init__(self, logger=None):
    """
    Initializes the one-time backup operation
    @param logger: The fwbackups logger to use. If none is supplied,
                   a new one will be created.
    """
    Common.__init__(self, logger)
    self.logger.logmsg('INFO', _('Starting restore operation'))
    self.config = config.RestoreConf()
    self.dest, self.source, self.sourceType, self.remoteHost, \
    self.remotePort, self.remoteUsername, self.remotePassword, \
    self.remoteFolder = self.getOptions(self.config)

  def getOptions(self, conf):
    """
    Loads all the configuration options from the .conf file
    `conf': The config.* instance to use
    """
    def makeBool(value):
      if value == 1 or value == '1' or value == True or value == 'True':
        return True
      else:
        return False
    destination = conf.get('Options', 'Destination')
    source = conf.get('Options', 'Source')
    sourceType = conf.get('Options', 'SourceType')
    remoteHost = conf.get('Options', 'RemoteHost')
    remotePort = conf.get('Options', 'RemotePort')
    if not remotePort:
      remotePort = 22
    else:
      remotePort = int(remotePort)
    remoteUsername = conf.get('Options', 'RemoteUsername')
    remotePassword = conf.get('Options', 'RemotePassword')
    remoteFolder = conf.get('Options', 'RemoteFolder')
    return destination, source, sourceType, remoteHost, remotePort, \
           remoteUsername, remotePassword, remoteFolder

  def restore(self):
    """ Restores a backup """
    wasErrors = False
    if self.sourceType == 'remote (ssh)': # check if server settings are OK
      self.logger.logmsg('DEBUG', _('Attempting to connect to server'))
      thread = fwbackups.runFuncAsThread(remote.testConnection, True,
                                         self.remoteHost, self.remoteUsername,
                                         self.remotePassword, self.remotePort,
                                         self.remoteFolder)
      while thread.retval == None:
        time.sleep(0.1)
      if thread.retval == True:
        pass
      elif thread.retval == False:
        self.logger.logmsg('ERROR', _('The selected file or folder was either not ' + \
                       'found or the user specified cannot write to it.'))
        return False
      elif thread.retval[0] == paramiko.AuthenticationException:
        self.logger.logmsg('ERROR', _('A connection was established, but authentication ' + \
                        'failed. Please verify the username and password ' + \
                        'and try again.'))
        return False
      elif thread.retval[0] == socket.gaierror or thread.retval[0] == socket.error:
        self.logger.logmsg('ERROR', _('A connection to the server could not be established.\n' + \
                        'Error %(a)s: %(b)s' % {'a': thread.retval[1][0], 'b': thread.retval[1][1]} + \
                        '\nPlease verify your settings and try again.'))
        return False
      elif thread.retval[0] == socket.timeout:
        self.logger.logmsg('ERROR', _('A connection to the server has timed out. ' + \
                        'Please verify your settings and try again.'))
        return False
      elif thread.retval[0] == paramiko.SSHException:
        self.logger.logmsg('ERROR', _('A connection to the server could not be established ' + \
                        'because an error occurred: %s' % thread.retval[1] + \
                        '\nPlease verify your settings and try again.'))
        return False
    # Set totals
    self.total = 1

    # source types:     'local set' 'local archive' 'local folder'
    #                   'remote archive (SSH)' 'remote folder (SSH)'

    # We don't want to raise a hard error, that's already in the log.
    # So we settle for a simple return false.
    # don't need source type logic, /destination/ is always a folder
    if not self.prepDestDir(self.dest):
     return False

    #FIXME: Remote folder not supported
    if self.sourceType == 'remote archive (SSH)' or self.sourceType == 'remote folder (SSH)':
      self.logger.logmsg('DEBUG', _('Receiving files from server via SFTP'))
      self.current += 1
      self.dest = os.path.join(self.dest, os.path.split(self.remoteFolder)[1])
      try:
        remote.receiveFile(self.remoteHost, self.remoteUsername, self.remotePassword, self.remotePort,
                           self.remoteFolder, self.dest)
      except Exception, error:
        self.logger.logmsg('DEBUG', _('Could not receive file from server: %s' % error))
        wasErrors = True

    self.current = 1
    try:

      if self.sourceType == 'local set': # we don't know the type
        if os.path.isfile(self.source):
          fh = tarfile.open(self.source, 'r')
          fh.extractall(self.dest)
          fh.close()
        elif os.path.isdir(self.source): # we are dealing with rsync
          shutil_modded.copytree(self.source, self.dest)
        else: # oops, something is up
          self.logger.logmsg('ERROR', _('Source `%s\' is not a file or folder!' % self.source))
          return False

      elif self.sourceType == 'local archive' or self.sourceType == 'remote archive (SSH)': # archives
        if os.path.isfile(self.source):
          fh = tarfile.open(self.source, 'r')
          fh.extractall(self.dest)
          fh.close()
        else: # oops, something is up
          self.logger.logmsg('ERROR', _('Source `%s\' is not an archive!' % self.source))
          return False

      else: # folders
        if not os.path.isdir(self.source):
          self.logger.logmsg('ERROR', _('Source `%s\' is not a folder' % self.source))
          return False
        if not self.prepDestDir(self.dest):
          return False
        shutil_modded.copytree(self.source, self.dest)

    except:
      self.logger.logmsg('ERROR','Error(s) occurred while restoring certain files or folders.\nPlease check the traceback below to determine if any files are incomplete or missing.')
      (etype, value, tb) = sys.exc_info()
      tracebackText = ''.join(traceback.format_exception(etype, value, tb))
      self.logger.logmsg('ERROR', _('Traceback: %s' % tracebackText))
      wasErrors = True

    # All done!
    self.logger.logmsg('INFO', _('Finished restore operation'))
    return not wasErrors

class Backup(Common):
  """ A wrapper for all backups """
  def __init__(self, logger=None):
    """
    Inititalize the generic backup operation
    @param logger: The fwbackups logger to use. If not specified,
                   a new one will be created.
    """
    Common.__init__(self, logger)

  def getOptions(self, conf):
    """
    Loads all the configuration options from the .conf file
    `conf': The config.* instance to use
    """
    def makeBool(value):
      if value == 1 or value == '1' or value == True or value == 'True':
        return True
      else:
        return False
    folderdest = conf.get('Options', 'Destination')
    engine = conf.get('Options', 'Engine')
    recursive = makeBool(conf.get('Options', 'Recursive'))
    pkglist = makeBool(conf.get('Options', 'PkgListsToFile'))
    diskinfo = makeBool(conf.get('Options', 'DiskInfoToFile'))
    dohidden = makeBool(conf.get('Options', 'BackupHidden'))
    excludes = conf.get('Options', 'Excludes')
    nice = int(conf.get('Options', 'Nice'))
    followlinks = makeBool(conf.get('Options', 'FollowLinks'))
    sparse = makeBool(conf.get('Options', 'Sparse'))
    destinationType = conf.get('Options', 'DestinationType')
    remoteHost = conf.get('Options', 'RemoteHost')
    remotePort = conf.get('Options', 'RemotePort')
    incremental = makeBool(conf.get('Options', 'Incremental'))
    if not remotePort:
      remotePort = 22
    else:
      remotePort = int(remotePort)
    remoteUsername = conf.get('Options', 'RemoteUsername')
    remotePassword = conf.get('Options', 'RemotePassword')
    remoteFolder = conf.get('Options', 'RemoteFolder')
    return folderdest, engine, recursive, pkglist, diskinfo,\
           dohidden, excludes, nice, sparse, followlinks,\
           destinationType, remoteHost, remotePort, remoteUsername,\
           remotePassword, remoteFolder, incremental

  def parsePaths(self, config):
    """
    Get the list of paths
    @returns: a list of paths to backup and the length of the paths
    """
    paths = []
    for pathEntry in config.options('Paths'):
      path = config.get('Paths', pathEntry)
      # non recursive - Skip dirs
      if self.recursive == False and os.path.isdir(path):
        continue # move on to next file
      paths.append("'%s'" % '\'\\\'\''.join(path.split('\''))) # wrap it in quotes for command line, and escape other single quote
    if paths == []:
      self.logger.logmsg('WARNING', _('After parsing options, there are no paths to backup!') + ' paths == []')
      return False, -1
    return paths, len(paths)

  def createPkgLists(self):
    """ Create the pkg lists in /tmp """
    managers = []
    for path in os.environ['PATH'].split(':'):
      if os.path.exists(os.path.join(path, 'rpm')):
        try: # try with rpm-python
          import rpm
          fullList = open('/tmp/%s.txt' % _('RPM - Package list, names only'), 'a')
          nameList = open('/tmp/%s.txt' % _('RPM - Package list'), 'a')
          ts=rpm.ts()
          # Equivalent to rpm -qa
          mi=ts.dbMatch()
          for hdr in mi:
            fullList.write("%s-%s-%s.%s" % (hdr['name'], hdr['version'], hdr['release'], hdr['arch']))
            nameList.write("%s.%s" % (hdr['name'], hdr['arch']))
          nameList.close()
          fullList.close()
        except: # otherwise just execute it like the rest
          fwbackups.execute('rpm -qa --qf "%%{NAME}\n" > "/tmp/%s.txt"' % _('RPM - Package list, names only'), env=self.environment, shell=True)
          fwbackups.execute('rpm -qa > "/tmp/%s.txt"' % _('RPM - Package list'), env=self.environment, shell=True)
        managers.append('RPM')
      if os.path.exists(os.path.join(path, 'pacman')):
        fwbackups.execute('pacman -Q > "/tmp/%s.txt"' % _('Pacman - Package list, names only'), env=self.environment, shell=True)
        fwbackups.execute('pacman -Qq > "/tmp/%s.txt"' % _('Pacman - Package list'), env=self.environment, shell=True)
        managers.append('Pacman')
      if os.path.exists(os.path.join(path, 'dpkg')):
        # FIXME: Package names only?
        fwbackups.execute('dpkg -l > "/tmp/%s.txt"' % _('dpkg - Package list'), env=self.environment, shell=True)
        managers.append('dpkg')
    return managers
       
  def createDiskInfo(self):
    """ Create the disk info in /tmp """
    fwbackups.execute('fdisk -l &> "/tmp/%s.txt"' % _('Disk Information'), env=self.environment, shell=True)

  def parseCommand(self, config, dest):
    """ Parse options to retrieve the correct command """
    folderdest, engine, recursive, pkglist, diskinfo,\
    dohidden, excludes, nice, sparse, followlinks,\
    destinationType, remoteHost, remotePort, remoteUsername,\
    remotePassword, remoteFolder, incremental = self.getOptions(config)
    if self.destinationType == 'remote (ssh)':
      import tempfile
      tempDir = tempfile.gettempdir()
      self.dest = os.path.join(tempDir, os.path.split(self.dest.replace("'", "'\\''"))[1])
    if self.engine == 'rsync':
      command = 'rsync -g -o -p -t -R'
      if incremental == True:
        command += ' -u --del'
      if recursive == True:
        command += ' -r'
      if not dohidden == True:
        command += ' --exclude=.*'
      if sparse == True:
        command += ' -S'
      if followlinks == True:
        command += ' -L'
      else:
        command += ' -l'
      if excludes != None and excludes != "":
        for i in excludes.split('\n'):
          command += ' --exclude="%s"' % i
    elif engine == 'tar':
      command = 'tar rf \'%s\'' % (dest.replace("'", "'\\''"))
    elif self.engine == 'tar.gz':
      # DON'T rfz - Can't use r (append) and z (gzip) together
      command = 'tar cfz \'%s\'' % (dest.replace("'", "'\\''"))
    elif self.engine == 'tar.bz2':
      # DON'T rfz - Can't use r (append) and z (gzip) together
      command = 'tar cfj \'%s\'' % (dest.replace("'", "'\\''"))
    # --
    if engine in ['tar', 'tar.gz', 'tar.bz2']: # they share command options
      if not recursive == True:
        command += ' --no-recursion'
      if not dohidden == True:
        command += ' --exclude=.*'
      if sparse == True:
        command += ' -S'
      if followlinks == True:
        command += ' -h'
    if excludes != None and excludes != "":
      for i in self.excludes.split('\n'):
        command += ' --exclude="%s"' % i

    return command

  def addListFilesToBackup(self, manager, command, engine, pkglist, diskinfo, paths):
    """  Adds the pkglist and diskinfo to the backup """
    self.ifCancel()
    if engine == 'tar':
      if pkglist == True:
        fwbackups.execute("%s '/tmp/%s - %s.txt' '/tmp/%s - %s.txt' '%s'" % \
          (command, manager, _('Package list'), manager, _('Package list, names only'),
          self.dest.replace("'", "'\\''")), env=self.environment, shell=True, timeout=-1)
      if diskinfo == True:
        fwbackups.execute("%s '/tmp/%s - %s.txt' '%s'" % \
          (command, manager, _('Disk Information'), self.dest.replace("'", "'\\''")), env=self.environment,
          shell=True, timeout=-1)
    elif engine in ['tar.gz', 'tar.bz2']:
      if pkglist == True:
        paths.append('"/tmp/%s - %s.txt" "/tmp/%s - %s.txt"' % \
        (manager, _('Package list'), manager, _('Package list, names only')))
      if diskinfo == True:
        paths.append('"/tmp/%s.txt"' % (_('Disk Information')))
    elif engine == 'rsync':
      if pkglist == True:
        fwbackups.execute("%s '/tmp/%s - %s.txt' '/tmp/%s - %s.txt' '%s'" % \
          (command, manager, _('Package list'), manager,  _('Package list, names only'),
          self.dest.replace("'", "'\\''")), env=self.environment, shell=True, timeout=-1)
      if diskinfo == True:
        # .replace("'", "'\\''") = wrap it in quotes for command line, and escape other single quote)
        fwbackups.execute("%s '/tmp/%s.txt' '%s'" % (command, _('Disk Information'), self.dest.replace("'", "'\\''")), env=self.environment, shell=True, timeout=-1)

  def deleteListFiles(self, manager, pkglist, diskinfo):
    """ Delete the list files in /tmp """
    if self.pkglist == True:
      os.remove('/tmp/%s - %s.txt' % (manager, _('Package list')) )
      os.remove('/tmp/%s - %s.txt' % (manager, _('Package list, names only')) )
    if self.diskinfo == True:
      os.remove('/tmp/%s.txt' % _('Disk Information'))
    self.ifCancel()

  def backupPaths(self, paths, command):
    """ Does the actual copying dirty work"""
    # this is in common
    wasAnError = False
    if self.destinationType == 'remote (ssh)': # check if server settings are OK
      self.logger.logmsg('DEBUG', _('Attempting to connect to server'))
      thread = fwbackups.runFuncAsThread(remote.testConnection, True,
                                         self.remoteHost, self.remoteUsername,
                                         self.remotePassword, self.remotePort,
                                         self.remoteFolder)
      while thread.retval == None:
        time.sleep(0.1)
      if thread.retval == True:
        pass
      elif thread.retval == False:
        self.logger.logmsg('ERROR', _('The selected file or folder was either not ' + \
                       'found or the user specified cannot write to it.'))
        return False
      elif thread.retval[0] == paramiko.AuthenticationException:
        self.logger.logmsg('ERROR', _('A connection was established, but authentication ' + \
                        'failed. Please verify the username and password ' + \
                        'and try again.'))
        return False
      elif thread.retval[0] == socket.gaierror or thread.retval[0] == socket.error:
        self.logger.logmsg('ERROR', _('A connection to the server could not be established.\n' + \
                        'Error %(a)s: %(b)s' % {'a': thread.retval[1][0], 'b': thread.retval[1][1]} + \
                        '\nPlease verify your settings and try again.'))
        return False
      elif thread.retval[0] == socket.timeout:
        self.logger.logmsg('ERROR', _('A connection to the server has timed out. ' + \
                        'Please verify your settings and try again.'))
        return False
      elif thread.retval[0] == paramiko.SSHException:
        self.logger.logmsg('ERROR', _('A connection to the server could not be established ' + \
                        'because an error occurred: %s' % thread.retval[1] + \
                        '\nPlease verify your settings and try again.'))
        return False
    self.current = 1
    if self.engine == 'tar':
      if MSWINDOWS:
        fh = tarfile.open(self.dest, 'w')
        for i in paths:
          self.ifCancel()
          # let's deal with real paths
          i = i.strip("'")
          self.logger.logmsg('INFO', _('Using %s on Windows: Cancel function will only take effect after a path has been completed.' % self.engine))
          self.logger.logmsg('DEBUG', _('Backing up path %(a)i/%(b)i: %(c)s' % {'a': self.current, 'b': self.total, 'c': i}))
          fh.add(i, recursive=self.recursive)
          self.logger.logmsg('DEBUG', _('Adding path `%s\' to the archive' % i))
          self.current += 1
        fh.close()
      else:
        for i in paths:
          self.ifCancel()
          self.logger.logmsg('DEBUG', _('Running command: nice -n %(a)i %(b)s %(c)s' % {'a': self.nice, 'b': command, 'c': i}))
          sub = fwbackups.executeSub('nice -n %i %s %s' % (self.nice, command, i), env=self.environment, shell=True)
          self.pids.append(sub.pid)
          self.logger.logmsg('DEBUG', _('Starting subprocess with PID %s') % sub.pid)
          # Sleep while not done.
          while sub.poll() in ["", None]:
            time.sleep(0.01)
          self.pids.remove(sub.pid)
          retval = sub.poll()
          self.logger.logmsg('DEBUG', _('Subprocess with PID %(a)s exited with status %(b)s' % {'a': sub.pid, 'b': retval}))
          # Something wrong?
          if retval != EXIT_STATUS_OK and retval != 2:
            wasAnError = True
            self.logger.logmsg('ERROR','Error(s) occurred while backing up path \'%s\'.\nPlease check the error output below to determine if any files are incomplete or missing.' % str(i))
            self.logger.logmsg('ERROR', _('Output:\n%s') % ('Return value: %s\nstdout: %s\nstderr: %s' % (retval, ''.join(sub.stdout.readlines()), ''.join(sub.stderr.readlines()) )))
          self.current += 1
      
    elif self.engine == 'tar.gz':
      self.total = 1
      if MSWINDOWS:
        fh = tarfile.open(self.dest, 'w:gz')
        for i in paths:
          self.ifCancel()
          # let's deal with real paths
          i = i.strip("'")
          self.logger.logmsg('INFO', _('Using %s on Windows: Cancel function will only take effect after a path has been completed.' % self.engine))
          self.logger.logmsg('DEBUG', _('Backing up path %(a)i/%(b)i: %(c)s' % {'a': self.current, 'b': self.total, 'c': i}))
          fh.add(i, recursive=self.recursive)
          self.logger.logmsg('DEBUG', _('Adding path `%s\' to the archive' % i))
          self.current += 1
        fh.close()
      else:
        i = ' '.join(paths)
        self.logger.logmsg('INFO', _('Using %s: Must backup all paths at once - Progress notification will be disabled.' % self.engine))
        self.logger.logmsg('DEBUG', _('Backing up path %(a)i/%(b)i: %(c)s') % {'a': self.current, 'b': self.total, 'c': i.replace("'", '')})
        self.logger.logmsg('DEBUG', _('Running command: nice -n %(a)i %(b)s %(c)s' % {'a': self.nice, 'b': command, 'c': i}))
        sub = fwbackups.executeSub('nice -n %i %s %s' % (self.nice, command, i), env=self.environment, shell=True)
        self.pids.append(sub.pid)
        self.logger.logmsg('DEBUG', _('Starting subprocess with PID %s') % sub.pid)
        # Sleep while not done.
        while sub.poll() in ["", None]:
          time.sleep(0.01)
        self.pids.remove(sub.pid)
        retval = sub.poll()
        self.logger.logmsg('DEBUG', _('Subprocess with PID %(a)s exited with status %(b)s' % {'a': sub.pid, 'b': retval}))
        # Something wrong?
        if retval != EXIT_STATUS_OK and retval != 2:
          wasAnError = True
          self.logger.logmsg('ERROR','Error(s) occurred while backing up path \'%s\'.\nPlease check the error output below to determine if any files are incomplete or missing.' % str(i))
          self.logger.logmsg('ERROR', _('Output:\n%s') % ('Return value: %s\nstdout: %s\nstderr: %s' % (retval, ''.join(sub.stdout.readlines()), ''.join(sub.stderr.readlines()) )))
          
    elif self.engine == 'tar.bz2':
      self.total = 1
      if MSWINDOWS:
        fh = tarfile.open(self.dest, 'w:bz2')
        for i in paths:
          self.ifCancel()
          # let's deal with real paths
          i = i.strip("'")
          self.logger.logmsg('INFO', _('Using %s on Windows: Cancel function will only take effect after a path has been completed.' % self.engine))
          self.logger.logmsg('DEBUG', _('Backing up path %(a)i/%(b)i: %(c)s' % {'a': self.current, 'b': self.total, 'c': i}))
          fh.add(i, recursive=self.recursive)
          self.logger.logmsg('DEBUG', _('Adding path `%s\' to the archive' % i))
          self.current += 1
        fh.close()
      else:
        i = ' '.join(paths)
        self.logger.logmsg('INFO', _('Using %s: Must backup all paths at once - Progress notification will be disabled.' % self.engine))
        self.logger.logmsg('DEBUG', _('Backing up path %(a)i/%(b)i: %(c)s') % {'a': self.current, 'b': self.total, 'c': i.strip('\'')[1]})
        self.logger.logmsg('DEBUG', _('Running command: nice -n %(a)i %(b)s %(c)s' % {'a': self.nice, 'b': command, 'c': i}))
        sub = fwbackups.executeSub('nice -n %i %s %s' % (self.nice, command, i), env=self.environment, shell=True)
        self.pids.append(sub.pid)
        self.logger.logmsg('DEBUG', _('Starting subprocess with PID %s') % sub.pid)
        # Sleep while not done.
        while sub.poll() in ["", None]:
          time.sleep(0.01)
        self.pids.remove(sub.pid)
        retval = sub.poll()
        self.logger.logmsg('DEBUG', _('Subprocess with PID %(a)s exited with status %(b)s' % {'a': sub.pid, 'b': retval}))
        # Something wrong?
        if retval != EXIT_STATUS_OK and retval != 2:
          wasAnError = True
          self.logger.logmsg('ERROR','Error(s) occurred while backing up path \'%s\'.\nPlease check the error output below to determine if any files are incomplete or missing.' % str(i))
          self.logger.logmsg('ERROR', _('Output:\n%s') % ('Return value: %s\nstdout: %s\nstderr: %s' % (retval, ''.join(sub.stdout.readlines()), ''.join(sub.stderr.readlines()) )))
          
    elif self.engine == 'rsync':
      # in this case, self.{folderdest,dest} both need to be created
      self.prepDestDir(self.dest)
      for i in paths:
        self.ifCancel()
        if MSWINDOWS:
          # let's deal with real paths
          i = i.strip("'")
          self.logger.logmsg('DEBUG', _('Backing up path %(a)i/%(b)i: %(c)s' % {'a': self.current, 'b': self.total, 'c': i}))
          shutil_modded.copytree_fullpaths(i, self.dest)
          self.current += 1
        else:
          self.logger.logmsg('DEBUG', _("Running command: nice -n %(a)i %(b)s %(c)s '%(d)s'" % {'a': self.nice, 'b': command, 'c': i, 'd': self.dest.replace("'", "'\\''")}))
          sub = fwbackups.executeSub("nice -n %i %s %s '%s'" % (self.nice, command, i, self.dest.replace("'", "'\\''")), env=self.environment, shell=True)
          self.pids.append(sub.pid)
          self.logger.logmsg('DEBUG', _('Starting subprocess with PID %s') % sub.pid)
          # Sleep while not done.
          while sub.poll() in ["", None]:
            time.sleep(0.01)
          self.pids.remove(sub.pid)
          retval = sub.poll()
          self.logger.logmsg('DEBUG', _('Subprocess with PID %(a)s exited with status %(b)s' % {'a': sub.pid, 'b': retval}))
          # Something wrong?
          if retval != EXIT_STATUS_OK and retval != 2:
            wasAnError = True
            self.logger.logmsg('ERROR','Error(s) occurred while backing up path \'%s\'.\nPlease check the error output below to determine if any files are incomplete or missing.' % str(i))
            self.logger.logmsg('ERROR', _('Output:\n%s') % ('Return value: %s\nstdout: %s\nstderr: %s' % (retval, ''.join(sub.stdout.readlines()), ''.join(sub.stderr.readlines()) )))
          self.current += 1
    # finally, we do this
    self.current = self.total
    time.sleep(1)
    self.ifCancel()
    if self.destinationType == 'remote (ssh)':
      self.logger.logmsg('DEBUG', _('Sending files to server via SFTP'))
      self.current += 1
      try:
        if os.path.isfile(self.dest):
          remote.transferFile(self.remoteHost, self.remoteUsername, self.remotePassword, self.remotePort,
                              self.dest, os.path.join(self.remoteFolder, os.path.split(self.dest)[1]), symlinks=not self.followlinks)
          os.remove(self.dest)
        elif os.path.isdir(self.dest):
          remote.transferFolder(self.remoteHost, self.remoteUsername, self.remotePassword, self.remotePort,
                                self.dest, os.path.join(self.remoteFolder, os.path.split(self.dest)[1]), symlinks=not self.followlinks)
          shutil_modded.rmtree(self.dest)
        else:
          wasAnError = True
          self.logger.logmsg('ERROR', _('Could not send file(s) or folder to server: Item(s) to transfer are not files or folders!'))
          self.logger.logmsg('DEBUG', _('Attempted to transfer: %s' % self.dest))
      except:
        wasAnError = True
        self.logger.logmsg('DEBUG', _('Could not send file(s) or folder to server:'))
        (etype, value, tb) = sys.exc_info()
        self.logger.logmsg('DEBUG', ''.join(traceback.format_exception(etype, value, tb)))
    return (not wasAnError)

class OneTime(Backup):
  """ The one-time backup operation """
  def __init__(self, logger=None):
    """ Initializes the one-time backup operation
    `logger': The fwbackups logger to use. New logger will be created if needed.
    """
    Backup.__init__(self, logger)
    self.logger.logmsg('INFO', _('Starting one-time backup operation'))
    self.config = config.OneTimeConf()
    self.folderdest, self.engine, self.recursive, self.pkglist, self.diskinfo,\
    self.dohidden, self.excludes, self.nice, self.sparse, self.followlinks,\
    self.destinationType, self.remoteHost, self.remotePort, self.remoteUsername,\
    self.remotePassword, self.remoteFolder, self.incremental = self.getOptions(self.config)
    # Parse backup folder format
    date = time.strftime('%Y-%m-%d_%H-%M')
    self.dest = ConvertPath('%s/%s-%s-%s' % (self.folderdest, _('Backup'), _('OneTime'), date))
    # IF tar || tar.gz, add .tar || tar.gz respectively to the dest since
    # the dest is to be a file, not a folder...
    if self.engine == 'tar':
      self.dest += '.tar'
    elif self.engine == 'tar.gz':
      self.dest += '.tar.gz'
    elif self.engine == 'tar.bz2':
      self.dest += '.tar.bz2'

  def backup(self):
    """ One-time backup"""
    paths, self.total = self.parsePaths(self.config)
    if not paths:
      return False

    if self.pkglist == True:
      managers = self.createPkgLists()
    else:
      managers = []
    if self.diskinfo == True:
      self.createDiskInfo()

    if not self.prepDestDir(self.folderdest):
      return False
    self.removeIfExists(self.dest)
    self.ifCancel()

    command = self.parseCommand(self.config, self.dest)
    for manager in managers:
      self.addListFilesToBackup(manager, command, self.engine, self.pkglist, self.diskinfo, paths)

    # Now that the paths & commands are set up...
    errorBool = self.backupPaths(paths, command)

    for manager in managers:
      self.deleteListFiles(manager, self.pkglist, self.diskinfo)
    self.ifCancel()

    if self.destinationType == 'local':
      try:
        os.chmod(self.dest, 0711)
      except:
        pass

    # All done!
    self.logger.logmsg('INFO', _('Finished one-time backup'))
    return errorBool



class Automated(Backup):
  """ Automated backup operation """
  def __init__(self, setname, logger=None):
    """
    Initializes the automatic backup operation
    setname: Name of the set to automatically backup
    logger (optional): The fwbackups logger to use. If none is supplied,
               a new one will be created.
    """
    Backup.__init__(self, logger)
    self.setConf = config.BackupSetConf(setname)
    self.logger.logmsg('INFO', _('Starting automatic backup operation of set `%s\'') % self.setConf.getSetName())
    self.folderdest, self.engine, self.recursive, self.pkglist, self.diskinfo,\
    self.dohidden, self.excludes, self.nice, self.sparse, self.followlinks,\
    self.destinationType, self.remoteHost, self.remotePort, self.remoteUsername,\
    self.remotePassword, self.remoteFolder, self.incremental = self.getOptions(self.setConf)
    # format destination
    date = time.strftime('%Y-%m-%d_%H-%M')
    self.dest = ConvertPath('%s/%s-%s-%s' % (self.folderdest, _('Backup'), self.setConf.getSetName(), date))
    # set-specific options
    self.command_before = self.setConf.get('Options', 'CommandBefore')
    self.command_after = self.setConf.get('Options', 'CommandAfter')
    if self.setConf.get('Options', 'Enabled') == '0':
      self.logger.logmsg('INFO', _('Set \'%s\' is disabled, will not continue') % self.setConf.getSetName())
      return
    self.tokeep = int(float(self.setConf.get('Options', 'OldToKeep')))
    # IF tar || tar.gz, add .tar || tar.gz respectively to the dest since
    # the dest is to be a file, not a folder...
    if self.engine == 'tar':
      self.dest += '.tar'
    elif self.engine == 'tar.gz':
      self.dest += '.tar.gz'
    elif self.engine == 'tar.bz2':
      self.dest += '.tar.bz2'
    
    #self.backup()

  def removeOldBackups(self):
    """ Get list of old backups and remove them"""
    listing = os.listdir(self.folderdest)
    listing.sort()
    oldbackups = []
    for i in listing:
      if i.startswith('%s-%s-' % (_('Backup'), self.setConf.getSetName())):
        oldbackups.append(i)
    # ...And remove them.
    oldbackups.reverse()
    if self.engine == 'rsync' and self.incremental == True and oldbackups:
      for i in oldbackups[:-1]:
        self.logger.logmsg('DEBUG', _('Removing old backup `%s\'') % i)
        shutil_modded.rmtree(path=ConvertPath('%s/%s' % (self.folderdest, i)), onerror=self.onError)
      oldIncrementalBackup = ConvertPath('%s/%s' % (self.folderdest, oldbackups[-1]))
      if not oldIncrementalBackup.endswith('.tar') and not oldIncrementalBackup.endswith('.tar.gz') and \
         not oldIncrementalBackup.endswith('.tar.bz2'): # oldIncrementalBackup = rsync
        self.logger.logmsg('DEBUG', _('Moving  `%s\' to `%s\'') % (oldIncrementalBackup, self.dest))
        shutil_modded.move(oldIncrementalBackup, self.dest)
      else: # source = is not a rsync backup - remove it and start fresh
        self.logger.logmsg('DEBUG', _('`%s\' is not an rsync backup - removing.') % oldIncrementalBackup)
        shutil_modded.rmtree(path=oldIncrementalBackup, onerror=self.onError)
    else:
      for i in oldbackups[self.tokeep:]:
        self.logger.logmsg('DEBUG', _('Removing old backup `%s\'') % i)
        shutil_modded.rmtree(path=ConvertPath('%s/%s' % (self.folderdest, i)), onerror=self.onError)
    


  def backup(self):
    """ Backup a set """
    # Get the list of paths...
    paths, self.total = self.parsePaths(self.setConf)
    if not paths:
      return False

    if not self.prepDestDir(self.folderdest):
      return False
    self.removeIfExists(self.dest)

    self.ifCancel()
    # Remove old stuff
    self.removeOldBackups()
    self.ifCancel()

    # Before command...
    if self.command_before:
      self.logger.logmsg('INFO', _('Executing \'Before\' command'))
      sub = fwbackups.executeSub(self.command_before, env=self.environment, shell=True)
      self.pids.append(sub.pid)
      self.logger.logmsg('DEBUG', _('Starting subprocess with PID %s') % sub.pid)
      # Sleep while not done.
      while sub.poll() in ["", None]:
        time.sleep(0.01)
      self.pids.remove(sub.pid)
      retval = sub.poll()
      # Something wrong?
      if retval != EXIT_STATUS_OK:
        self.logger.logmsg('ERROR', _('Command returned with a non-zero exit status:'))
        self.logger.logmsg('ERROR', 'Return value: %s\nstdout: %s\nstderr: %s' % (retval, ''.join(sub.stdout.readlines()), ''.join(sub.stderr.readlines()) ))
      self.ifCancel()
    
    if self.pkglist == True:
      managers = self.createPkgLists()
    else:
      managers = []
    if self.diskinfo == True:
      self.createDiskInfo()
    self.ifCancel()

    command = self.parseCommand(self.setConf, self.dest)
    for manager in managers:
      self.addListFilesToBackup(manager, command, self.engine, self.pkglist, self.diskinfo, paths)
    # Now that the paths & commands are set up...
    errorBool = self.backupPaths(paths, command)
    for manager in managers:
      self.deleteListFiles(manager, self.pkglist, self.diskinfo)
    if self.destinationType == 'local':
      try:
        os.chmod(self.dest, 0711)
      except:
        pass

    # After command
    if self.command_after:
      self.logger.logmsg('INFO', _('Executing \'After\' command'))
      sub = fwbackups.executeSub(self.command_after, env=self.environment, shell=True)
      self.pids.append(sub.pid)
      self.logger.logmsg('DEBUG', _('Starting subprocess with PID %s') % sub.pid)
      # Sleep while not done.
      while sub.poll() in ["", None]:
        time.sleep(0.01)
      self.pids.remove(sub.pid)
      retval = sub.poll()
      # Something wrong?
      if retval != EXIT_STATUS_OK:
        self.logger.logmsg('ERROR', _('Command returned with a non-zero exit status:'))
        self.logger.logmsg('ERROR', 'Return value: %s\nstdout: %s\nstderr: %s' % (retval, ''.join(sub.stdout.readlines()), ''.join(sub.stderr.readlines()) ))
    self.ifCancel()

    # All done!
    self.logger.logmsg('INFO', _('Finished automatic backup operation of set \'%s\'') % self.setConf.getSetName())
    return errorBool

